#include "NetworkWindow.h"

#include <ScrollView.h>
#include <Box.h>
#include <stdlib.h>


#include "GoWindow.h"
#include "GameItem.h"

using namespace muscle;

BMessageTransceiverThread gMuscle;

NetworkWindow::NetworkWindow()
:	BWindow(
		BRect(50,50,450,350),
		"Internet Go",
		B_TITLED_WINDOW,
		B_ASYNCHRONOUS_CONTROLS|B_QUIT_ON_WINDOW_CLOSE|B_NOT_RESIZABLE
	)
{
	gMuscle.SetTarget( BMessenger(this) );
	
	BView * background = new BView( Bounds(), "background", B_FOLLOW_ALL, 0 );
	background->SetViewColor( 218,218,218 );
	
	AddChild( background );
	
	// new game
	BMessage * nameMsg = new BMessage('Name');
	
	fNewGameText = new BTextControl(
		BRect(10,10,200,30),
		"GameName",
		"Game name", "Name your game",
		NULL
	);
	fNewGameText->SetDivider( 100 );
	
	fNewGameButton = new BButton(
		BRect(210,10,300,40),
		"NewGameButton",
		"Create game",
		new BMessage(*nameMsg)
	);
	fNewGameButton->SetEnabled(false);
	
	background->AddChild( fNewGameText );
	background->AddChild( fNewGameButton );
	
	// join game
	
	BBox * joinBox = new BBox(
		BRect(10,70,390,290), "Join game"
	);
	
	BMessage * joinMsg = new BMessage('Join');
	
	fGameList = new BListView(
		BRect(10,20,270,210), "GameList",
		B_SINGLE_SELECTION_LIST
	);
	fGameList->SetInvocationMessage( joinMsg );
	
	BScrollView * sv = new BScrollView(
		"", fGameList, B_FOLLOW_NONE, 0, false, true
	);
	
	fJoinGameButton = new BButton(
		BRect(310,70,370,110),
		"JoinGameButton",
		"Join game",
		new BMessage(*joinMsg)
	);
	fJoinGameButton->SetEnabled(false);
	
	joinBox->AddChild( sv );
	joinBox->AddChild( fJoinGameButton );
	
	background->AddChild( joinBox );
}

NetworkWindow::~NetworkWindow()
{
}

string next_session_id("a");

void
NetworkWindow::MessageReceived( BMessage * msg )
{
	switch ( msg->what )
	{
		case 'Remo':
		{ // a GoWindow has been closed, remove from fWindows
			map<string,GoWindow*>::iterator i;
			i = fWindows.find( msg->FindString("go_session") );
			
			if ( i != fWindows.end() )
			{
				fWindows.erase(i);
			}
			
		}	break;
		
		case 'Name':
		{ // Create new game
			printf("Create game\n");
			
			MessageRef msg;
			
			// Generate session ID
			string sessionID(next_session_id);
			next_session_id += "a";
			
			string session_path = "Go/Session/" + sessionID;
			
			string local_addr = fLocalAddress;// + "/" + session_path;
			
			// Create window
			Go * go = new Go(9,9);
			
			GoWindow * win = new GoWindow(
				this,
				go,
				fNewGameText->Text(),
				NULL, 
				local_addr.c_str(), 
				sessionID.c_str() 
			);
			win->Show();
			
			// set session <> window mapping
			fWindows[sessionID] = win;
			
			// set session message
			Message sessionMsg('Make');
			
			msg = GetMessageFromPool(PR_COMMAND_SETDATA);
			msg()->AddMessage( session_path.c_str(), sessionMsg );
			
			gMuscle.SendMessageToSessions(msg);
			
			// publish advertising message
			string publish_path = string("Go/Publish/") + sessionID;
			
			Message goMsg('Publ');
			goMsg.AddString("game name", fNewGameText->Text() );
			goMsg.AddString("go_session", sessionID.c_str() );
			goMsg.AddString("address", local_addr.c_str() );
			
			msg = GetMessageFromPool(PR_COMMAND_SETDATA);
			msg()->AddMessage( publish_path.c_str(), goMsg );
			
			gMuscle.SendMessageToSessions(msg);
		}	break;
		/*
		
		Go/Games/<new name>
		
		*/
		case 'Join':
		{ // Join a game!
			printf("Join game\n");
			
			GameItem * item = (GameItem*)fGameList->ItemAt( 
				fGameList->CurrentSelection() 
			);
			
			/*
				Create GoWindow.
				Add it to address,window map
				Run it.
			*/
			
			Go * go = new Go(9,9);
			
			GoWindow * win = new GoWindow(
				this,
				go, 
				item->Text(),
				item->fAddress.c_str(),
				fLocalAddress.c_str(),
				item->fSession.c_str()
			);
			win->Show();
			
			fWindows[item->fSession] = win;
			
		}	break;
		
		case muscle::MUSCLE_THREAD_SIGNAL:
		{
			MessageRef msg;
			uint32 code;
			while(gMuscle.GetNextEventFromInternalThread(code, &msg) >= 0)
			{
				switch(code)
				{
					case MTT_EVENT_INCOMING_MESSAGE:
					{
						printf("Incoming message!\n");
						Message * pMsg = msg();  // Get access to the reference's held Message object.
						//pMsg->PrintToStream();
						HandleMessage(pMsg);     // Do whatever you gotta do
						/* do NOT delete (pMsg).  It will be deleted for you. */
					}	break;
					
					case MTT_EVENT_SESSION_CONNECTED:
					{
						printf("Connection to server was successful!\n");
			         	
						// subscribe to games
						MessageRef msg = GetMessageFromPool(PR_COMMAND_SETPARAMETERS);
						msg()->AddBool("SUBSCRIBE:Go/Publish/*", true);
						gMuscle.SendMessageToSessions(msg);
						
						msg = GetMessageFromPool(PR_COMMAND_GETPARAMETERS);
						gMuscle.SendMessageToSessions(msg);
					} break;
					
					case MTT_EVENT_SESSION_DISCONNECTED:
						printf("Disconnected from server, or connection failed!\n");
						break;
				}
			}
		}	break;
		default:
			BWindow::MessageReceived(msg);
			break;
	}
}

void
NetworkWindow::HandleMessage( Message * msg )
{
	switch ( msg->what )
	{
		case PR_RESULT_PARAMETERS:
		{
			const char * addr = NULL;
			msg->FindString("!Root", &addr);
			fLocalAddress = addr;
			printf("Local address: %s\n", fLocalAddress.c_str() );
			
			fNewGameButton->SetEnabled(true);
			fJoinGameButton->SetEnabled(true);
		}	break;
		
		case PR_RESULT_DATAITEMS:
		{
			printf("Matching items:\n");
			
			MessageFieldNameIterator iter = msg->GetFieldNameIterator();
			
			const String * field;
			const char * removed;
			
			for ( int i=0; msg->FindString(PR_NAME_REMOVED_DATAITEMS,i,&removed)==B_OK; i++ )
			{ // remove any needed stuff
				printf("Removing game: %s\n", removed );
				
				for ( int x=0; x<fGameList->CountItems(); x++ )
				{
					GameItem * item = (GameItem*)fGameList->ItemAt(x);
					
					string path = item->fAddress + "/Go/Publish/" + item->fSession;
					
					if ( path == removed )
					{
						fGameList->RemoveItem(item);
						delete item;
					}
				}
			}
			
			while ( (field = iter.GetNextFieldNameString()) != NULL )
			{
				printf("Field: %s\n", field->Cstr());
				
				Message publ;
				
				msg->FindMessage( *field, publ );
				
				if ( publ.what == 'Publ' )
				{ // something's published, add to list
					printf("Adding new game to list\n");
					const char * name=NULL;
					const char * addr=NULL;
					const char * session=NULL;
					
					publ.FindString("game name", &name );
					publ.FindString("address", &addr );
					publ.FindString("go_session", &session );
					
					GameItem * item = new GameItem( name );
					item->fAddress = addr;
					item->fSession = session;
					
					fGameList->AddItem( item );
				}
			}
		}	break;
		
		default:
		{
			const char * session = NULL;
			
			if ( msg->FindString("go_session", &session) == B_OK )
			{
				printf("Got a message for go_session '%s'\n", session);
				map<string,GoWindow*>::iterator i = fWindows.find(session);
				
				if ( i != fWindows.end() )
				{
					fWindows[session]->HandleGoMessage(msg);
				} else
				{
					printf("Unknown session\n");
					msg->PrintToStream();
				}
			} else
			{
				printf("Unknown message!\n");
				msg->PrintToStream();
			}
		}	break;
	}
}
